package worldGraph;

import java.util.ArrayList;
import java.util.Stack;
import agent.Agent;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardWorldModel;
import geometry.Mathematic;
import geometry.Point;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import utilities.Logger;

public class WorldGraph {
	public ArrayList<Area> areas = null;

	// This variable shows the Enterances which there is a way from our position
	// to that enterances
	public ArrayList<Enterance> myEnterances = null;
	public Area myArea = null;

	public static int RAD = 500;
	private static int MIN_EDGE_LENGTH = 500;

	private StandardWorldModel model = null;

	public WorldGraph(StandardWorldModel model,
			ArrayList<Building> modelBuildings, ArrayList<Road> modelRoads,
			ArrayList<Refuge> modelRefuges) {
		this.model = model;
		areas = new ArrayList<worldGraph.Area>();

		ArrayList<rescuecore2.standard.entities.Area> modelAreas = new ArrayList<rescuecore2.standard.entities.Area>();
		modelAreas.addAll(modelBuildings);
		modelAreas.addAll(modelRoads);
		modelAreas.addAll(modelRefuges);

		for (rescuecore2.standard.entities.Area area : modelAreas) {
			Area newArea = new Area(area);
			areas.add(newArea);
		}

		for (Area area : areas) {
			rescuecore2.standard.entities.Area modelArea = area.modelArea;
			for (Edge edge : modelArea.getEdges()) {
				area.points.add(new Point(edge.getStartX(), edge.getStartY()));
				if (edge.getNeighbour() != null
						&& Mathematic.getDistance(edge.getStartX(), edge
								.getStartY(), edge.getEndX(), edge.getEndY()) > MIN_EDGE_LENGTH)
					area.enterances.add(new Enterance(area.enterances.size(),
							area.points.size() - 1, area, RAD,
							(rescuecore2.standard.entities.Area) model
									.getEntity(edge.getNeighbour())));
			}
			area.points.add(new Point(modelArea.getEdges().get(0).getStartX(),
					modelArea.getEdges().get(0).getStartY()));

			area.updateCenterPoint();
			if (area.enterances.size() == 0)
				System.out.println("Map BUG: Area: "
						+ area.modelArea.getID().getValue());
			for (Enterance enterance : area.enterances)
				enterance.createPoints();
		}

		for (Area area : areas) {
			int areaID = area.modelArea.getID().getValue();
			for (Enterance enterance : area.enterances) {
				Area otherArea = enterance.neighbourAreaInModel.worldGraphArea;
				for (Enterance neighbour : otherArea.enterances)
					if (neighbour.neighbourAreaInModel.getID().getValue() == areaID) {
						enterance.neighbour = neighbour;
						break;
					}
			}
		}

		for (Area area : areas)
			area.updateAvailablePoints();

		for (Area area : areas)
			area.updateEnterancesConnectivity(RAD, null, false);
	}

	public void clearAreas() {
		for (Area area : areas)
			area.clearMarks();
	}

	public boolean isThereWayFromHere(Area target) {
		if (myEnterances == null)
			return false;
		if (myArea != null && myArea == target)
			return true;
		setTarget(target);

		return isThereWay(myEnterances);
	}

	public boolean isThereWay(EntityID start, EntityID target) {
		return isThereWay(
				((rescuecore2.standard.entities.Area) model.getEntity(start)).worldGraphArea,
				((rescuecore2.standard.entities.Area) model.getEntity(target)).worldGraphArea);
	}

	public boolean isThereWay(Area start, Area target) {
		if (start == target)
			return true;
		setTarget(target);

		return isThereWay(start.enterances);
	}

	public void clearTargets() {
		for (Area area : areas)
			area.setAreaAsTarget(false);
	}

	public void setTarget(Area target) {
		clearTargets();
		target.setAreaAsTarget(true);
	}

	public boolean isThereWay(ArrayList<Enterance> start) {
		clearAreas();

		ArrayList<Enterance> layer = new ArrayList<Enterance>();
		for (Enterance enterance : start)
			if (!enterance.mark) {
				enterance.mark = true;
				layer.add(enterance);
			}

		while (layer.size() > 0) {
			ArrayList<Enterance> newLayer = new ArrayList<Enterance>();
			for (Enterance enterance : layer) {
				if (enterance.isItConnectedToNeighbour
						&& !enterance.neighbour.mark) {
					Enterance neighbour = enterance.neighbour;
					if (neighbour.isTarget)
						return true;

					neighbour.mark = true;
					for (Enterance internal : neighbour.internalEnterances) {
						if (!internal.mark) {
							if (internal.isTarget)
								return true;

							internal.mark = true;
							newLayer.add(internal);
						}
					}
				}
			}

			layer = newLayer;
		}

		return false;
	}

	public ArrayList<EntityID> getMinPath(EntityID start, EntityID end,
			boolean checkBlockades) {
		return getMinPath(
				((rescuecore2.standard.entities.Area) model.getEntity(start)).worldGraphArea,
				((rescuecore2.standard.entities.Area) model.getEntity(end)).worldGraphArea,
				checkBlockades);
	}

	public ArrayList<EntityID> getMinPath(Area start, Area end,
			boolean checkBlockades) {
		return getMinPath(start.enterances, end, checkBlockades);
	}

	public ArrayList<EntityID> getMinPath(ArrayList<Enterance> start, Area end,
			boolean checkBlockades) {
		setTarget(end);
		return getMinPath(start, checkBlockades);
	}

	public ArrayList<EntityID> getMinPath(ArrayList<Enterance> start,
			boolean checkBlockades) {
		clearAreas();
		ArrayList<EntityID> path = new ArrayList<EntityID>();
		ArrayList<Enterance> layer = new ArrayList<Enterance>();
		for (Enterance enterance : start)
			if (enterance.isTarget)
				return path;
			else if (!checkBlockades || !enterance.mark) {
				enterance.mark = true;
				layer.add(enterance);
			}

		Enterance lastEnterance = null;
		done: while (layer.size() > 0) {
			ArrayList<Enterance> newLayer = new ArrayList<Enterance>();
			for (Enterance enterance : layer)
				if (!checkBlockades
						|| (enterance.isItConnectedToNeighbour && !enterance.neighbour.mark)) {
					Enterance neighbour = enterance.neighbour;
					neighbour.lastEnterance = enterance;
					if (neighbour.isTarget) {
						lastEnterance = neighbour;
						break done;
					}

					neighbour.mark = true;
					if (checkBlockades) {
						for (Enterance internal : neighbour.internalEnterances)
							if (!internal.mark) {
								internal.lastEnterance = neighbour;
								if (internal.isTarget) {
									lastEnterance = internal;
									break done;
								}

								internal.mark = true;
								newLayer.add(internal);
							}
					} else {
						for (Enterance internal : neighbour.area.enterances)
							if (neighbour != internal && internal.mark == false) {
								internal.lastEnterance = neighbour;
								if (internal.isTarget) {
									lastEnterance = internal;
									break done;
								}

								internal.mark = true;
								newLayer.add(internal);
							}
					}
				}

			layer = newLayer;
		}

		Stack<EntityID> stack = new Stack<EntityID>();
		while (lastEnterance != null) {
			EntityID entityID = lastEnterance.area.modelArea.getID();
			if (stack.size() == 0 || stack.peek() != entityID)
				stack.add(entityID);
			lastEnterance = lastEnterance.lastEnterance;
		}

		while (stack.size() != 0)
			path.add(stack.pop());

		return path;
	}

	public boolean log = false;
	public int time = -1;

	public void update(ChangeSet changeSet) {
		for (EntityID entityID : changeSet.getChangedEntities()) {
			Entity e = model.getEntity(entityID);
			if (e instanceof rescuecore2.standard.entities.Road) {
				rescuecore2.standard.entities.Area modelArea = (rescuecore2.standard.entities.Area) e;
				Area area = modelArea.worldGraphArea;
				area.blockades.clear();
				if (modelArea.isBlockadesDefined())
					for (EntityID blockadeID : modelArea.getBlockades()) {
						rescuecore2.standard.entities.Blockade blockade = (rescuecore2.standard.entities.Blockade) model
								.getEntity(blockadeID);
						area.blockades.add(new Blockade(blockade));
					}
				area.updateAvailablePoints();
			}
		}

		logger.log("after update availablePoints");
		for (EntityID entityID : changeSet.getChangedEntities()) {
			Entity e = model.getEntity(entityID);
			if (e instanceof rescuecore2.standard.entities.Area)
				((rescuecore2.standard.entities.Area) e).worldGraphArea
						.updateEnterancesConnectivity(RAD, changeSet, true);
		}
	}

	@SuppressWarnings("unchecked")
	private Agent agent = null;

	public Logger logger = null;

	@SuppressWarnings("unchecked")
	public void update(int time, Agent agent, EntityID myPosition, int myX,
			int myY) {
		this.time = time;
		this.agent = agent;

		if (myPosition != null && myX != -1 && myY != -1) {
			myArea = ((rescuecore2.standard.entities.Area) model
					.getEntity(myPosition)).worldGraphArea;
			myEnterances = myArea.updateEnterancesConnectivity(new Point(myX,
					myY), RAD, agent.changeSet, true);
		}
		if (myEnterances == null)
			myEnterances = new ArrayList<Enterance>();
	}

	@SuppressWarnings("unchecked")
	public void addReachableArea(Enterance enterance) {
		boolean oldValue = enterance.area.isReachable;
		enterance.area.isReachable = true;
		rescuecore2.standard.entities.Area area = enterance.area.modelArea;
//		agent.reachableEnterances.put(enterance.index, enterance);

//		if (!agent.reachableAreas.contains(area)) {
		if (!oldValue) {
//			agent.reachableAreas.put(area.getID().getValue(), area);
			area.worldGraphArea.distanceFromSelf = distance;
			agent.reachableAreas.add(area);

			if (area instanceof Refuge)
				agent.reachableRefuges.add((Refuge) area);
			else if (area instanceof Building)
				agent.reachableBuildings.add((Building) area);
			else if (area instanceof Road)
				agent.reachableRoads.add((Road) area);
		}
	}

	private int distance = -1;
	@SuppressWarnings("unchecked")
	public void updateReachableAreas(ArrayList<Enterance> start) {
		agent.reachableBuildings = new ArrayList<Building>();
		agent.reachableRoads = new ArrayList<Road>();
		agent.reachableRefuges = new ArrayList<Refuge>();
//		agent.reachableAreas = new TreeMap<Integer, rescuecore2.standard.entities.Area>();
//		agent.reachableEnterances = new TreeMap<Integer, Enterance>();
		agent.reachableAreas = new ArrayList<rescuecore2.standard.entities.Area>();
//		agent.reachableAreas = new HashSet<rescuecore2.standard.entities.Area>();

		clearAreas();
		for (Area area : areas) {
			area.distanceFromSelf = -1;
			area.isReachable = false;
			area.buildingReachability = false;
			area.buildingReachabilityCalculated = false;
		}

		distance = 0;
		ArrayList<Enterance> layer = new ArrayList<Enterance>();
		for (Enterance enterance : start)
			if (!enterance.mark) {
				addReachableArea(enterance);
				enterance.mark = true;
				layer.add(enterance);
			}

		while (layer.size() > 0) {
			ArrayList<Enterance> newLayer = new ArrayList<Enterance>();
			distance++;
			for (Enterance enterance : layer) {
				if (enterance.isItConnectedToNeighbour
						&& !enterance.neighbour.mark) {
					Enterance neighbour = enterance.neighbour;
					addReachableArea(neighbour);
					neighbour.mark = true;
					for (Enterance internal : neighbour.internalEnterances) {
						if (!internal.mark) {
							addReachableArea(internal);
							internal.mark = true;
							newLayer.add(internal);
						}
					}
				}
			}

			layer = newLayer;
		}
	}

	// Ambulance getMinPath
	public ArrayList<Civilian> getMinPathForAmbulance(ArrayList<Enterance> start) {
		clearAreas();
		ArrayList<Enterance> layer = new ArrayList<Enterance>();
		for (Enterance enterance : start)
			if (enterance.isTarget)
				return enterance.civilians;
			else if (!enterance.mark) {
				enterance.mark = true;
				layer.add(enterance);
			}

		while (layer.size() > 0) {
			ArrayList<Enterance> newLayer = new ArrayList<Enterance>();
			for (Enterance enterance : layer)
				if (enterance.isItConnectedToNeighbour
						&& !enterance.neighbour.mark) {
					Enterance neighbour = enterance.neighbour;
					if (neighbour.isTarget)
						return neighbour.civilians;

					neighbour.mark = true;
					for (Enterance internal : neighbour.internalEnterances)
						if (!internal.mark) {
							if (internal.isTarget)
								return internal.civilians;

							internal.mark = true;
							newLayer.add(internal);
						}
				}

			layer = newLayer;
		}

		return null;
	}
}
